unit IOCPHandler;

interface

uses Windows, Winsock2, Classes, SysUtils, Dialogs, SyncObjs, TypesDefs;

const
  MAX_WORKER_THREAD = 16;

  SEND_RESULT_DONE  = 0;
  SEND_RESULT_DCED  = -1;

  MSG_FREE_CLIENT_CONTEXT              = 0;
  MSG_ERROR_SEND_NIL_CONTEXT           = 1;
  MSG_ERROR_CONNECTED_SEND_NIL_CONTEXT = 2;

type
    TDynByteArray = array of byte;

    PIO_OPERATION = ^IO_OPERATION;
    IO_OPERATION = (ClientIoRead,ClientIoWrite);

    // data to be associated for every I/O operation on a socket
    PPER_IO_CONTEXT = ^PER_IO_CONTEXT;
    PER_IO_CONTEXT = packed record
      m_Overlapped: WSAOVERLAPPED;
      wsa_buf: WSABUF;
      IOOperation: IO_OPERATION;
      SocketAccept: TSocket;
    end;

    // data to be associated with every socket added to the IOCP
    PPER_SOCKET_CONTEXT = ^PER_SOCKET_CONTEXT;
    PER_SOCKET_CONTEXT = packed record
      Socket: TSocket;
      pIOReadContext,pIOWriteContext: PPER_IO_CONTEXT;
    end;

    PClientContextBySocket = ^TClientContextBySocket;
    TClientContextBySocket = array[0..maxword]of PPER_SOCKET_CONTEXT;


    TOnDataReadEvent = procedure (ppsContext: PPER_SOCKET_CONTEXT; count: dword);
    TOnDataWriteEvent = procedure (ppsContext: PPER_SOCKET_CONTEXT; transferred: dword);
    TOnAcceptEvent = procedure (ppsContext: PPER_SOCKET_CONTEXT);
    TOnDisconnectEvent = procedure (ppsContext: PPER_SOCKET_CONTEXT);
    TOnSetBufferEvent = procedure (ppsContext: PPER_SOCKET_CONTEXT);

    TCustomSocket = class
    protected
      FIsOpen: boolean;
      FCompletionPort: THandle;
      FOwnSocket: TSocket;
      FEndServer: boolean;
      FCS: TCriticalSection;
      FThreadCount: dword;              //worker thread count
      FThreadHandles: array[0..MAX_WORKER_THREAD-1]of THandle; //TWOHandleArray???
      FCtxtList: TClientContextBySocket;
      function ContextListAllocate(sd: TSocket):PPER_SOCKET_CONTEXT;
      procedure ContextListAdd(ppsContext: PPER_SOCKET_CONTEXT);
      procedure ContextListRemove(ppsContext: PPER_SOCKET_CONTEXT; operation: IO_OPERATION);
      procedure ContextListFree;
    public
      OnClientRead: TOnDataReadEvent;
      OnClientWrite: TOnDataWriteEvent;
      OnClientAccept: TOnAcceptEvent;
      OnClientDisconnect: TOnDisconnectEvent;
      property IsOpen: boolean read FIsOpen;
      property EndServer: boolean read FEndServer;
      property CompletionPort: THandle read FCompletionPort;
      constructor Create(threadCount: dword);
      destructor Destroy;override;
      procedure CloseContextSocket(socket: dword; operation: IO_OPERATION; bGraceful: boolean = true);
      function Send(socket: TSocket; Buffer: pointer; len: dword):boolean;
    end;

    TListenSocket = class(TCustomSocket)
    private
      FListenAddress: sockaddr_in;
      FAcceptThread: THandle;
      function CreateListenSocket: boolean;
    protected
    public
      constructor Create(const ListenAddress: string; Port: word;
        OnDataReadEvent: TOnDataReadEvent = nil; OnDataWriteEvent: TOnDataWriteEvent = nil;
        OnAcceptEvent: TOnAcceptEvent = nil; OnDisconnectEvent: TOnDisconnectEvent = nil);
      //destructor Destroy;override;
    end;

    TConnectedSocket = class(TCustomSocket)
    private
      FDestAddress: sockaddr_in;
      function CreateConnectedSocket: boolean;
    protected
    public
      constructor Create(const DestAddress: string; Port: word; OnSetBufferEvent: TOnSetBufferEvent;
        OnDataReadEvent: TOnDataReadEvent = nil; OnDataWriteEvent: TOnDataWriteEvent = nil;
        OnDisconnectEvent: TOnDisconnectEvent = nil);
      //destructor Destroy;override;
      property CSocket: TSocket read FOwnSocket;
    end;

    function ScktErrorCheck(scktError: integer):integer;

implementation

uses LogHandler;

function ScktErrorCheck(scktError: integer):integer;
begin
     result:=scktError;
     if scktError = SOCKET_ERROR then
        AddCustomLog(WSAGetLastError, 0, SOCKET_MESSAGELOG);
end;

function ScktErrorCheckPending(scktError: integer):integer;
var
   lastError: integer;
begin
     result:=scktError;
     if scktError = SOCKET_ERROR then begin
        lastError:=WSAGetLastError;
        if lastError <> WSA_IO_PENDING then begin
           AddCustomLog(lastError, 0, SOCKET_MESSAGELOG);
        end
        else result:=0;
     end;
end;

{ WorkerThread }

function WorkerThread(Owner: TListenSocket): dword;stdcall;
var
   hIOCP: THandle;
   bSuccess: boolean;
   nRet: integer;
   //ppsContext: PPER_SOCKET_CONTEXT;
   lpIOContext: PPER_IO_CONTEXT;
   dwRecvNumBytes,dwFlags,dwIoSize,sckt: dword;
begin
     hIOCP:= Owner.CompletionPort;
     lpIOContext:=nil;
     dwRecvNumBytes:= 0;
     //dwSendNumBytes:= 0;
     dwFlags:= 0;
     result:=0;

     while true do begin
		       bSuccess:= GetQueuedCompletionStatus(hIOCP,dwIoSize,sckt,POverlapped(lpIOContext),INFINITE); //lpOverlapped

           if not(bSuccess) then ErrorCheck;

           if sckt = 0 then
              // PostQueuedCompletionStatus to post an I/O packet with
              // a NULL CompletionKey (or if we get one for any reason).  It is time to exit.
              exit;
           //socket:=ppsContext.Socket;
           if Owner.EndServer then
              // main thread will do all cleanup needed - see finally block
              exit;

           if not(bSuccess) or (bSuccess and (dwIoSize = 0)) then begin
              // client connection dropped, continue to service remaining (and possibly
              // new) client connections
              Owner.CloseContextSocket(sckt,lpIOContext.IOOperation,false);
              continue;
           end;

           // determine what type of IO packet has completed by checking the PER_IO_CONTEXT
           // associated with this socket.  This will determine what action to take.
           if lpIOContext.IOOperation = ClientIoRead then begin
              if Assigned(Owner.OnClientRead) then
              try
                 Owner.OnClientRead(Owner.FCtxtList[sckt],dwIoSize); //can be destroyed
              except on e: Exception do AddToLog(e.Message);
              end;
              if (sckt > high(Owner.FCtxtList))or(Owner.FCtxtList[sckt] = nil) then continue;

              ZeroMemory(lpIOContext,sizeof(OVERLAPPED));
				      dwFlags:= 0;
              //   
				      nRet:= WSARecv(sckt, @lpIOContext.wsa_buf, 1, dwRecvNumBytes, dwFlags, PWSAOverlapped(@lpIOContext.m_Overlapped), nil);
				      if ScktErrorCheckPending(nRet) = SOCKET_ERROR then
                 Owner.CloseContextSocket(sckt,ClientIoRead,false);
           end
           else if lpIOContext.IOOperation = ClientIoWrite then begin
              ZeroMemory(lpIOContext,sizeof(OVERLAPPED));
              if Assigned(Owner.OnClientWrite) then
              try
                 Owner.OnClientWrite(Owner.FCtxtList[sckt],dwIoSize);
              except on e: Exception do AddToLog(e.Message);
              end;
           end;
     end;
end;

{ TAcceptThread }

procedure AcceptConnection(Owner: TListenSocket; ClientSocket: TSocket);
var
   nRet: integer;
   dwFlags,dwBytes: dword;
   ppsContext: PPER_SOCKET_CONTEXT;
begin
     ppsContext:= Owner.ContextListAllocate(ClientSocket);
     if ppsContext = nil then exit;

     if CreateIoCompletionPort(ClientSocket, Owner.CompletionPort, ClientSocket{dword(ppsContext)}, 0) = 0 then begin
        ErrorCheck;
        Owner.ContextListRemove(ppsContext,ClientIoRead);
        exit;
     end;

     Owner.ContextListAdd(ppsContext);

     if Assigned(Owner.OnClientAccept) then
        Owner.OnClientAccept(ppsContext)
     else exit;                                                //wsabuf not init
     if ppsContext.pIOReadContext.wsa_buf.buf = nil then exit; //wsabuf not set

		 dwFlags:= 0;
		 nRet:= WSARecv(ClientSocket, @ppsContext.pIOReadContext.wsa_buf, 1, dwBytes, dwFlags, PWSAOverlapped(@ppsContext.pIOReadContext.m_Overlapped), nil);
		 if ScktErrorCheckPending(nRet) = SOCKET_ERROR then
		    //Error occurred while executing WSARecv()
        Owner.CloseContextSocket(ClientSocket,ClientIoRead,false);
end;

function AcceptThread(Owner: TListenSocket):dword;stdcall;
var
   ClientAddress: TSockAddrIn;
   addrlen: integer;
   s: TSocket;
begin
	   //Accept thread will be around to look for accept event, until a Shutdown event is not Signaled.
     addrlen:=sizeof(TSockAddrIn);
     while not(Owner.EndServer) do begin
           s := WSAAccept(Owner.FOwnSocket, @ClientAddress, @addrlen, nil, 0);
           if s = INVALID_SOCKET then begin
              ScktErrorCheck(s);
           end
           else AcceptConnection(Owner,s);   //(' ...');
	         //Display Client's IP
	         //WriteToConsole("\nClient connected from: %s", inet_ntoa(ClientAddress.sin_addr));
     end;
     result:=0;
end;

{ TCustomSocket }

constructor TCustomSocket.Create(threadCount: dword);
var
   hThread,ThreadID: dword;
   i: dword;
begin
     FIsOpen:=false;
     FEndServer:=false;
     FCS:=TCriticalSection.Create;
     FThreadCount:=threadCount;

     FCompletionPort:=CreateIoCompletionPort(INVALID_HANDLE_VALUE,0,0,FThreadCount);
     if FCompletionPort = 0 then begin
        AddToLog('CreateIoCompletionPort failed to create I/O completion port');
        ErrorCheck;
        FThreadCount:=0;exit;
     end;

     for i:= 0 to FThreadCount-1 do begin
         hThread:=CreateThread(nil, 0, @WorkerThread, self, 0, ThreadID);
         if hThread = 0 then begin
            AddToLog('CreateThread failed to create worker thread');
            ErrorCheck;
            exit;
         end;
		     FThreadHandles[i]:=hThread;
     end;

     ZeroMemory(@FCtxtList,sizeof(FCtxtList));
end;

destructor TCustomSocket.Destroy;
var
   i: dword;
begin
     FEndServer:= true;

     // Cause worker threads to exit
     if FCompletionPort <> 0 then
        for i:=0 to FThreadCount-1 do
            PostQueuedCompletionStatus(FCompletionPort, 0, 0, nil);

     //Make sure worker threads exits.
     if (WAIT_FAILED = WaitForMultipleObjects( FThreadCount,  @FThreadHandles, true, 1000)) then begin
         AddToLog('WaitForMultipleObjects failed');ErrorCheck;
     end
     else
        for i:=0 to FThreadCount-1 do begin
            if (FThreadHandles[i] <> INVALID_HANDLE_VALUE) then
               CloseHandle(FThreadHandles[i]);
            FThreadHandles[i]:= INVALID_HANDLE_VALUE;
        end;

     ContextListFree;

     if FCompletionPort <> 0 then begin
        CloseHandle(FCompletionPort);
        FCompletionPort:=0;
     end;

     if FOwnSocket <> INVALID_SOCKET then begin
        closesocket(FOwnSocket);
        FOwnSocket:= INVALID_SOCKET;
     end;

     FCS.Free;
     inherited;
end;

procedure TCustomSocket.CloseContextSocket(socket: dword; operation: IO_OPERATION; bGraceful: boolean);
var
    lingerStruct: TLinger;
begin
     if socket = 0 then exit;
     
     AddToLog('CloseContextSocket: '+IntToStr(socket));
     if Assigned(OnClientDisconnect)and(FCtxtList[socket] <> nil) then OnClientDisconnect(FCtxtList[socket]);
     FCS.Enter;
     if FCtxtList[socket] = nil then begin
        AddToLog('CloseClient: ppsContext is NULL');
        FCS.Leave;
        exit;
     end;
     if not(bGraceful) then begin
        // force the subsequent closesocket to be abortative.
        lingerStruct.l_onoff:= 1;
        lingerStruct.l_linger:= 0;
        ScktErrorCheck(SetSockOpt(socket, SOL_SOCKET, SO_LINGER, PChar(@lingerStruct), sizeof(lingerStruct)));
     end;
     CloseSocket(socket);

     ContextListRemove(FCtxtList[socket],operation);
     FCS.Leave;
end;

function TCustomSocket.ContextListAllocate(sd: TSocket):PPER_SOCKET_CONTEXT;
var
   ppsContext: PPER_SOCKET_CONTEXT;
begin
     result:=nil;
     FCS.Enter;
     ppsContext:=HeapAlloc(GetProcessHeap, 8{HEAP_ZERO_MEMORY},sizeof(PER_SOCKET_CONTEXT));
     if ppsContext = nil then begin
        AddToLog('HeapAlloc PER_SOCKET_CONTEXT Failed');
        ErrorCheck;
        FCS.Leave;
        exit;
     end;

     ppsContext.pIOReadContext:=HeapAlloc(GetProcessHeap, 8{HEAP_ZERO_MEMORY},sizeof(PER_IO_CONTEXT));
     ppsContext.pIOWriteContext:=HeapAlloc(GetProcessHeap, 8{HEAP_ZERO_MEMORY},sizeof(PER_IO_CONTEXT));
     if (ppsContext.pIOReadContext = nil)or(ppsContext.pIOWriteContext = nil) then begin
        HeapFree(GetProcessHeap, 0, ppsContext);
        AddToLog('HeapAlloc PER_IO_CONTEXT Failed');
        ErrorCheck;
        FCS.Leave;
        exit;
     end;

     ppsContext.Socket:= sd;

     ppsContext.pIOReadContext.IOOperation := ClientIoRead;

     ppsContext.pIOWriteContext.IOOperation := ClientIoWrite;

     FCS.Leave;

     result:=ppsContext;
end;

procedure TCustomSocket.ContextListAdd(ppsContext: PPER_SOCKET_CONTEXT);
begin
     FCS.Enter;
     if FCtxtList[ppsContext.Socket] <> nil then
        ContextListRemove(ppsContext,ClientIoRead);
     FCtxtList[ppsContext.Socket]:=ppsContext;
     FCS.Leave;
end;

procedure TCustomSocket.ContextListRemove(ppsContext: PPER_SOCKET_CONTEXT; operation: IO_OPERATION);
begin
     FCS.Enter;
     if (ppsContext = nil)or(ppsContext.Socket > high(FCtxtList))or(FCtxtList[ppsContext.Socket] = nil) then begin
        AddToLog('ContextListRemove: ppsContext is NULL');
        FCS.Leave;
        exit;
     end;
     if FCtxtList[ppsContext.Socket] <> nil then begin
        if (ppsContext.pIOReadContext <> nil)and(operation = ClientIoRead) then begin
            //The overlapped structure is safe to free when only the posted i/o has
            //completed. Here we only need to test those posted but not yet received
            //by PQCS in the shutdown process.
            if FEndServer then
                while ppsContext.pIOReadContext.m_Overlapped.Internal = STATUS_PENDING do
                      Sleep(0);
            HeapFree(GetProcessHeap, 0, ppsContext.pIOReadContext);
            ppsContext.pIOReadContext:=nil;
        end;
        if ppsContext.pIOWriteContext <> nil then begin
            //The overlapped structure is safe to free when only the posted i/o has
            //completed. Here we only need to test those posted but not yet received
            //by PQCS in the shutdown process.
            if FEndServer then
                while ppsContext.pIOWriteContext.m_Overlapped.Internal = STATUS_PENDING do
                      Sleep(0);
            HeapFree(GetProcessHeap, 0, ppsContext.pIOWriteContext);
            ppsContext.pIOWriteContext:=nil;
        end;
     end;

     if (ppsContext.pIOReadContext = nil)and(ppsContext.pIOWriteContext = nil)then begin
        FCtxtList[ppsContext.Socket]:=nil;
        HeapFree(GetProcessHeap, 0, ppsContext);
     end;
     FCS.Leave;
end;

procedure TCustomSocket.ContextListFree;
var
   i: integer;
begin
     FCS.Enter;
     for i:=0 to high(FCtxtList) do
        if FCtxtList[i] <> nil then        
          CloseContextSocket(FCtxtList[i].Socket,ClientIoRead,false);
     FCS.Leave;
end;

function TCustomSocket.Send(socket: TSocket; Buffer: pointer; len: dword):boolean;
var
   dwFlags,dwBytes: dword;
   nRet: integer;
   ppsContext: PPER_SOCKET_CONTEXT;
begin
     result:=true;
     FCS.Enter;
     ppsContext:=FCtxtList[socket];
     if ppsContext = nil then begin
        AddCustomLog(MSG_ERROR_SEND_NIL_CONTEXT, 0, MESSAGELOG);
        FCS.Leave;
        exit;
     end;

     ZeroMemory(@ppsContext.pIOWriteContext.m_Overlapped,sizeof(OVERLAPPED));
     ppsContext.pIOWriteContext.wsa_buf.len:= len;
     ppsContext.pIOWriteContext.wsa_buf.buf:= Buffer;
     FCS.Leave;

     //log part
     AddToLog('Send to socket: ' + IntToStr(socket));
     //AddCustomLog(dword(Buffer), len, SEND_BUF_LOG);

     dwFlags:= 0;
     nRet:= WSASend(socket, @ppsContext.pIOWriteContext.wsa_buf, 1, dwBytes, dwFlags, PWSAOverlapped(ppsContext.pIOWriteContext), nil);
     if ScktErrorCheckPending(nRet) = SOCKET_ERROR then
        CloseContextSocket(socket,ClientIoWrite);
end;

{ TListenSocket }

constructor TListenSocket.Create(const ListenAddress: string; Port: word;
      OnDataReadEvent: TOnDataReadEvent = nil; OnDataWriteEvent: TOnDataWriteEvent = nil;
      OnAcceptEvent: TOnAcceptEvent = nil; OnDisconnectEvent: TOnDisconnectEvent = nil);
var
   si: SYSTEM_INFO;
begin
     FListenAddress.sin_family:= AF_INET;
     FListenAddress.sin_port:= ntohs(Port);
     //FListenAddress.sin_addr.S_addr:= htonl(INADDR_ANY);
     FListenAddress.sin_addr.S_addr:=Inet_addr(PChar(ListenAddress));
     ZeroMemory(@FListenAddress.sin_zero,SizeOf(FListenAddress.sin_zero));

     GetSystemInfo(si);

     inherited Create(si.dwNumberOfProcessors * 2);

     if not(CreateListenSocket) then exit;

     OnClientRead:=OnDataReadEvent;
     OnClientWrite:=OnDataWriteEvent;
     OnClientAccept:=OnAcceptEvent;
     OnClientDisconnect:=OnDisconnectEvent;

     FIsOpen:=true;
end;

function TListenSocket.CreateListenSocket: boolean;
var
    lingerStruct: TLinger;
    option,nZero,ThreadID: dword;
begin
     result:=false;
     lingerStruct.l_onoff:= 1;
     lingerStruct.l_linger:= 0;

     FOwnSocket:= ScktErrorCheck(WSASocket(AF_INET, SOCK_STREAM, IPPROTO_TCP, nil, 0, WSA_FLAG_OVERLAPPED));
     if FOwnSocket = INVALID_SOCKET then exit;

     option:= 1;
     if ScktErrorCheck(SetSockOpt(FOwnSocket, SOL_SOCKET, SO_REUSEADDR, PChar(@option), sizeof(option))) = SOCKET_ERROR then exit;

     if ScktErrorCheck(Bind(FOwnSocket, @FListenAddress, sizeof(FListenAddress))) = SOCKET_ERROR then exit;

     if ScktErrorCheck(Listen(FOwnSocket,SOMaxConn)) = SOCKET_ERROR then exit;

     // Disable send buffering on the socket.  Setting SO_SNDBUF
     // to 0 causes winsock to stop bufferring sends and perform
     // sends directly from our buffers, thereby reducing CPU usage.
     nZero:= 0;
     if ScktErrorCheck(SetSockOpt(FOwnSocket, SOL_SOCKET, SO_SNDBUF, PChar(@nZero), sizeof(nZero))) = SOCKET_ERROR then exit;

     // Disable receive buffering on the socket.  Setting SO_RCVBUF
     // to 0 causes winsock to stop bufferring receive and perform
     // receives directly from our buffers, thereby reducing CPU usage.
     nZero:= 0;
     if ScktErrorCheck(SetSockOpt(FOwnSocket, SOL_SOCKET, SO_RCVBUF, PChar(@nZero), sizeof(nZero))) = SOCKET_ERROR then exit;

     if ScktErrorCheck(SetSockOpt(FOwnSocket, SOL_SOCKET, SO_LINGER,PChar(@lingerStruct), sizeof(lingerStruct))) = SOCKET_ERROR then exit;

	   FAcceptThread := CreateThread(nil, 0, @AcceptThread, self, 0, ThreadID);
     if FAcceptThread = 0 then exit;

     result:=true;
end;

{ TConnectedSocket }

constructor TConnectedSocket.Create(const DestAddress: string; Port: word; OnSetBufferEvent: TOnSetBufferEvent;
  OnDataReadEvent: TOnDataReadEvent; OnDataWriteEvent: TOnDataWriteEvent;
  OnDisconnectEvent: TOnDisconnectEvent);
var
   dwFlags,dwBytes: dword;
   nRet: integer;
begin
     FOwnSocket:=INVALID_SOCKET;

     FDestAddress.sin_family:= AF_INET;
     FDestAddress.sin_port:= ntohs(Port);
     FDestAddress.sin_addr.S_addr:=Inet_addr(PChar(DestAddress));
     ZeroMemory(@FDestAddress.sin_zero,SizeOf(FDestAddress.sin_zero));

     inherited Create(2);

     if not(CreateConnectedSocket) then exit;

     OnSetBufferEvent(FCtxtList[FOwnSocket]);

     OnClientRead:= OnDataReadEvent;
     OnClientWrite:= OnDataWriteEvent;
     OnClientDisconnect:= OnDisconnectEvent;


		 dwFlags:= 0;
		 nRet:= WSARecv(FOwnSocket, @FCtxtList[FOwnSocket].pIOReadContext.wsa_buf, 1, dwBytes, dwFlags, PWSAOverlapped(@FCtxtList[FOwnSocket].pIOReadContext.m_Overlapped), nil);
		 if ScktErrorCheckPending(nRet) = SOCKET_ERROR then begin
        ContextListRemove(FCtxtList[0],ClientIoRead);exit;
     end;

     FIsOpen:=true;
end;

function TConnectedSocket.CreateConnectedSocket: boolean;
var
    lingerStruct: TLinger;
    option,nZero: dword;
    ppsContext: PPER_SOCKET_CONTEXT;
begin
     result:=false;
     lingerStruct.l_onoff:= 1;
     lingerStruct.l_linger:= 0;

     FOwnSocket:= ScktErrorCheck(WSASocket(AF_INET, SOCK_STREAM, IPPROTO_TCP, nil, 0, WSA_FLAG_OVERLAPPED));
     if FOwnSocket = INVALID_SOCKET then exit;

     option:= 1;
     if ScktErrorCheck(SetSockOpt(FOwnSocket, SOL_SOCKET, SO_REUSEADDR, PChar(@option), sizeof(option))) = SOCKET_ERROR then exit;
     if ScktErrorCheck(Connect(FOwnSocket, @FDestAddress, sizeof(FDestAddress))) = SOCKET_ERROR then begin
        ErrorCheck;
        closesocket(FOwnSocket);
        exit;
     end;

     // Disable send buffering on the socket.  Setting SO_SNDBUF
     // to 0 causes winsock to stop bufferring sends and perform
     // sends directly from our buffers, thereby reducing CPU usage.
     nZero:= 0;
     if ScktErrorCheck(SetSockOpt(FOwnSocket, SOL_SOCKET, SO_SNDBUF, PChar(@nZero), sizeof(nZero))) = SOCKET_ERROR then exit;

     // Disable receive buffering on the socket.  Setting SO_RCVBUF
     // to 0 causes winsock to stop bufferring receive and perform
     // receives directly from our buffers, thereby reducing CPU usage.
     nZero:= 0;
     if ScktErrorCheck(SetSockOpt(FOwnSocket, SOL_SOCKET, SO_RCVBUF, PChar(@nZero), sizeof(nZero))) = SOCKET_ERROR then exit;

     if ScktErrorCheck(SetSockOpt(FOwnSocket, SOL_SOCKET, SO_LINGER,PChar(@lingerStruct), sizeof(lingerStruct))) = SOCKET_ERROR then exit;

     ppsContext:= ContextListAllocate(FOwnSocket);
     if ppsContext = nil then exit;

     ContextListAdd(ppsContext);

	   if CreateIoCompletionPort(FOwnSocket, FCompletionPort, FOwnSocket{dword(ppsContext)}, 0) = 0 then begin
        ErrorCheck;
        ContextListRemove(ppsContext,ClientIoRead);
        exit;
     end;

     result:=true;
end;

end.
