unit TypesDefs;

interface

uses Windows, Classes, StrUtils, AcedContainers, AcedAlgorithm, Winsock2;

const
  SIZE32      = 32;
  //DIGEST_SIZE = 20;

type
  DWORD = cardinal;
  //PDWord = ^ DWORD;

  TDynIntArray = array of integer;
  TStringArray = array of string;

  TWSABufArr = array of WSABUF;

  //PObject = ^TObject;

  TThreadedIntegerHashtable = class(TObject)
  private
    FLock: TRTLCriticalSection;
    FObject: TIntegerHashtable;
    function GetItem(Key: Integer): Pointer;
    procedure SetItem(Key: Integer; Value: Pointer);
  public
    property Items[Key: Integer]: Pointer read GetItem write SetItem; default;
    constructor Create(InitialCapacity: Integer = 0);
    destructor Destroy;override;
    function Lock: TIntegerHashtable;
    procedure UnLock;
    procedure Add(Key: Integer; Value: Pointer);
    function Contains(Key: Integer): Boolean;
    procedure Remove(Key: Integer);
  end;

  TThreadedInt64Hashtable = class(TObject)
  private
    FLock: TRTLCriticalSection;
    FObject: TInt64Hashtable;
    function GetItem(Key: Int64): Pointer;
    procedure SetItem(Key: Int64; Value: Pointer);
  public
    property Items[Key: Int64]: Pointer read GetItem write SetItem; default;
    constructor Create(InitialCapacity: Integer = 0);
    destructor Destroy;override;
    function Lock: TInt64Hashtable;
    procedure UnLock;
    procedure Add(Key: Int64; Value: Pointer);
    function Contains(Key: Int64): Boolean;
    procedure Remove(Key: Int64);
  end;

  TThreadedStringHashtable = class(TObject)
  private
    FLock: TRTLCriticalSection;
    FObject: TStringHashtable;
    function GetItem(const Key: string): Pointer;
    procedure SetItem(const Key: string; Value: Pointer);
  public
    property Items[const Key: string]: Pointer read GetItem write SetItem; default;
    constructor Create(CaseSensitive: Boolean; InitialCapacity: Integer = 0);
    destructor Destroy;override;
    function Lock: TStringHashtable;
    procedure UnLock;
    procedure Add(const Key: string; Value: Pointer);
    function Contains(const Key: string): Boolean;
    procedure Remove(const Key: string);
  end;

  TThreadedIntegerList = class(TObject)
  private
    FLock: TRTLCriticalSection;
    FObject: TIntegerList;
  public
    constructor Create(InitialCapacity: Integer = 0);
    destructor Destroy;override;
    function Lock: TIntegerList;
    procedure UnLock;
  end;

  TThreadedStringList = class(TObject)
  private
    FLock: TRTLCriticalSection;
    FObject: TStringList;
  public
    constructor Create;
    destructor Destroy;override;
    function Lock: TStringList;
    procedure UnLock;
  end;

  function bswap(v: dword): dword;
  function GetAbsTickDiff(start_tick: dword): integer;
  function CompareFunctionPointer(Item1, Item2: Pointer): Integer;
  //function GetObject(pobj: PObject): TObject;

implementation

function bswap(v: dword):dword;
asm
  bswap eax
end;

function GetAbsTickDiff(start_tick: dword): integer;
var
   curr_tick: dword;
begin
     curr_tick:= GetTickCount;
     if curr_tick >= start_tick then
        result:= curr_tick - start_tick
     else result:= maxdword - start_tick + curr_tick;
end;

{  ,      
  .     Item1  Item2   
   ,     ,   ,
      ,  ,   . }

{function GetObject(pobj: PObject): TObject;
begin
  result:= TObject(pobj^);
end;}

function CompareFunctionPointer(Item1, Item2: Pointer): Integer;
begin
     if integer(Item1) < integer(Item2) then result:=-1
     else if integer(Item1) > integer(Item2) then result:=1
     else result:=0;
{asm
   cmp eax,edx todo
   flag to eax}
end;


{ TThreadedIntegerHashtable }

constructor TThreadedIntegerHashtable.Create(InitialCapacity: Integer);
begin
     inherited Create;
     InitializeCriticalSectionAndSpinCount(FLock,100);
     FObject:=TIntegerHashtable.Create(InitialCapacity);
end;

destructor TThreadedIntegerHashtable.Destroy;
begin
     Lock;
     try
        FObject.Free;
        inherited Destroy;
     finally
        Unlock;
        DeleteCriticalSection(FLock);
     end;
end;

function TThreadedIntegerHashtable.Lock: TIntegerHashtable;
begin
     EnterCriticalSection(FLock);
     result:=FObject;
end;

procedure TThreadedIntegerHashtable.Unlock;
begin
     LeaveCriticalSection(FLock);
end;

function TThreadedIntegerHashtable.GetItem(Key: Integer): Pointer;
begin
     Lock;
     result:=FObject.Items[Key];
     Unlock;
end;

procedure TThreadedIntegerHashtable.SetItem(Key: Integer; Value: Pointer);
begin
     Lock;
     FObject.Items[Key]:=value;
     Unlock;
end;

procedure TThreadedIntegerHashtable.Add(Key: Integer; Value: Pointer);
begin
     Lock;
     try
        FObject.Add(Key,Value);
     finally Unlock;
     end;
end;

function TThreadedIntegerHashtable.Contains(Key: Integer): Boolean;
begin
     Lock;
     result:=FObject.Contains(Key);
     Unlock;
end;

procedure TThreadedIntegerHashtable.Remove(Key: Integer);
begin
     Lock;
     FObject.Remove(Key);
     Unlock;
end;

{ TThreadedInt64Hashtable }

constructor TThreadedInt64Hashtable.Create(InitialCapacity: Integer);
begin
     inherited Create;
     InitializeCriticalSectionAndSpinCount(FLock,100);
     FObject:=TInt64Hashtable.Create(InitialCapacity);
end;

destructor TThreadedInt64Hashtable.Destroy;
begin
     Lock;
     try
        FObject.Free;
        inherited Destroy;
     finally
        Unlock;
        DeleteCriticalSection(FLock);
     end;
end;

function TThreadedInt64Hashtable.Lock: TInt64Hashtable;
begin
     EnterCriticalSection(FLock);
     result:=FObject;
end;

procedure TThreadedInt64Hashtable.Unlock;
begin
     LeaveCriticalSection(FLock);
end;

function TThreadedInt64Hashtable.GetItem(Key: Int64): Pointer;
begin
     Lock;
     result:=FObject.Items[Key];
     Unlock;
end;

procedure TThreadedInt64Hashtable.SetItem(Key: Int64; Value: Pointer);
begin
     Lock;
     FObject.Items[Key]:=value;
     Unlock;
end;

procedure TThreadedInt64Hashtable.Add(Key: Int64; Value: Pointer);
begin
     Lock;
     try
        FObject.Add(Key,Value);
     finally Unlock;
     end;
end;

function TThreadedInt64Hashtable.Contains(Key: Int64): Boolean;
begin
     Lock;
     result:=FObject.Contains(Key);
     Unlock;
end;

procedure TThreadedInt64Hashtable.Remove(Key: Int64);
begin
     Lock;
     FObject.Remove(Key);
     Unlock;
end;

{ TThreadedStringHashtable }

constructor TThreadedStringHashtable.Create(CaseSensitive: Boolean;
  InitialCapacity: Integer);
begin
     inherited Create;
     InitializeCriticalSectionAndSpinCount(FLock,100);
     FObject:=TStringHashtable.Create(CaseSensitive,InitialCapacity);
end;

destructor TThreadedStringHashtable.Destroy;
begin
     Lock;
     try
        FObject.Free;
        inherited Destroy;
     finally
        Unlock;
        DeleteCriticalSection(FLock);
     end;
end;

function TThreadedStringHashtable.Lock: TStringHashtable;
begin
     EnterCriticalSection(FLock);
     result:=FObject;
end;

procedure TThreadedStringHashtable.Unlock;
begin
     LeaveCriticalSection(FLock);
end;

function TThreadedStringHashtable.GetItem(const Key: string): Pointer;
begin
     Lock;
     result:=FObject.Items[Key];
     Unlock;
end;

procedure TThreadedStringHashtable.SetItem(const Key: string;
  Value: Pointer);
begin
     Lock;
     FObject.Items[Key]:=value;
     Unlock;
end;

procedure TThreadedStringHashtable.Add(const Key: string; Value: Pointer);
begin
     Lock;
     try
        FObject.Add(Key,Value);
     finally Unlock;
     end;
end;

function TThreadedStringHashtable.Contains(const Key: string): Boolean;
begin
     Lock;
     result:=FObject.Contains(Key);
     Unlock;
end;

procedure TThreadedStringHashtable.Remove(const Key: string);
begin
     Lock;
     FObject.Remove(Key);
     Unlock;
end;

{ TThreadedIntegerList }

constructor TThreadedIntegerList.Create(InitialCapacity: Integer);
begin
     inherited Create;
     InitializeCriticalSectionAndSpinCount(FLock,100);
     FObject:=TIntegerList.Create(InitialCapacity);
end;

destructor TThreadedIntegerList.Destroy;
begin
    Lock;
    try
       FObject.Free;
       inherited Destroy;
    finally
       Unlock;
       DeleteCriticalSection(FLock);
    end;
    inherited;
end;

function TThreadedIntegerList.Lock: TIntegerList;
begin
     EnterCriticalSection(FLock);
     result:=FObject;
end;

procedure TThreadedIntegerList.UnLock;
begin
     LeaveCriticalSection(FLock);
end;

{ TThreadedStringList }

constructor TThreadedStringList.Create;
begin
    inherited Create;
    InitializeCriticalSectionAndSpinCount(FLock,100);
    FObject:= TStringList.Create;
end;

destructor TThreadedStringList.Destroy;
begin
    Lock;
    try
       FObject.Free;
       inherited Destroy;
    finally
       Unlock;
       DeleteCriticalSection(FLock);
    end;
    inherited;
end;

function TThreadedStringList.Lock: TStringList;
begin
     EnterCriticalSection(FLock);
     result:=FObject;
end;

procedure TThreadedStringList.UnLock;
begin
     LeaveCriticalSection(FLock);
end;

end.
